!*  <<< SUBIO >>> ------------- documents are incomplete --------------
!+
!----------------------------------------------------------------------
!======================================================================

!*MATRD    -- read a matrix (real*4) from the FITS-formatted file
!+
SUBROUTINE MATRD(IM,G,iext,jext,IRC)
  integer    :: group
  real(4)    :: nullval, G(iext,jext)
  logical    :: anyf
!
! Arguments:
!  IM    (input)     : logical unit number of input data file
!  G     (output)    : matrix G(iext,jext)
!  iext  (input)     : I-dimension of the matrix
!  jext  (input)     : J-dimension of the matrix
!  IRC   (output)    : return code  ( 0 : normal end )
!
!--
!     AUTHOR: M. Hamabe  1990-10-17
!     UPDATE: M. Hamabe  1991-09-10   (data-type)
!     UPDATE: M. Hamabe  1996-07-17   (support FITS)
!     UPDATE: M. Hamabe  1999-07-17   (omit  IMFORT)
! ---------------------------------------------------------------------

  group=0
  nelements=iext*jext
  nullval=0.0
  call FTGPVE(im,group,1,nelements,nullval,G,anyf,irc)
  
  return

end SUBROUTINE MATRD
!=======================================================================

!*MATWT    -- write a matrix (real*4) as an FITS-formatted file (real).
!+
SUBROUTINE MATWT(IM,G,iext,jext,IRC)
  
  real  :: G(iext*jext)

!
! Arguments:
!  IM    (input)     : logical unit number of output data file
!  iext  (input)     : I-dimension of the matrix
!  jext  (input)     : J-dimension of the matrix
!  IRC   (output)    : return code  ( 0 : normal end )
!
!--
!     AUTHOR: M.Hamabe 1990-10-17
!     UPDATE: M.Hamabe 1996-08-21 (support FITS)
!     UPDATE: M.Hamabe 1999-07-22 (omit  IMFORT)
! ---------------------------------------------------------------------

  nelements=iext*jext
  call FTPPRE(im,0,1,nelements,G,irc)
  
  return

end SUBROUTINE MATWT

!=======================================================================

!*MATWTI   -- write a matrix (real*4) as an FITS-formatted file (reversely).
!+
SUBROUTINE MATWTI(IM,G,iext,jext,PIX,IRC)
  real :: G(iext,jext), PIX(iext)
!
! MATWTI:
! Write a matrix (real*4) as an FITS-formatted file (real)
! Reverse order along J-direction.
!
! Arguments:
!  IM   (input)     : logical unit number of output data file
!  iext (input)     : I-dimension of the matrix
!  jext (input)     : J-dimension of the matrix
!  G    (input)     : matrix G(iext,jext)
!  PIX              : work area
!
!--
!     AUTHOR: M.Hamabe 1992-04-17
!     UPDATE: M.Hamabe 1996-08-26 (support FITS)
!     UPDATE: M.Hamabe 1999-07-17 (omit  IMFORT)
! ---------------------------------------------------------------------
  
  jh=jext/2
  do j=1, jh
     do i=1, iext
        pix(i)=g(i,jext-j+1)
     end do
     do i=1, iext
        g(i,jext-j+1)=g(i,j)
        g(i,j)=pix(i)
     end do
  end do
  call FTPPRE(im,0,1,iext*jext,g,irc)
  
  return

end SUBROUTINE MATWTI

!=======================================================================

!*MATWTIR  -- write a matrix (real) as an FITS-formatted file (for ruh88)
!+
SUBROUTINE MATWTIR(IM,G,iext,jext,PIX,IRC)
  real :: G(iext,jext), PIX(jext)

! MATWTIR:
!         Write a matrix (real*4) as an FITS-formatted file (real).
!         Reverse order along J-direction.
!         Rotate 90 degree.
!         for ruh88 task.
! Arguments:
!           im     : LOGICAL UNIT NUMBER OF OUTPUT DATA FILE
!           iext   : I-dimension OF THE MATRIX
!           jext   : J-dimension OF THE MATRIX
!           g      : MATRIX G(iext,jext)
!           pix    : WORK AREA
!
!--
!     AUTHOR: M.Hamabe 1992-05-18
!     UPDATE: M.Hamabe 1996-08-27  (support FITS)
!     UPDATE: M.Hamabe 1999-07-17  (omit  IMFORT)
! ---------------------------------------------------------------------
!
  do i=1,iext
     do j=1, jext
        pix(j)=g(iext-i+1,j)
     end do
     call FTPPRE(im,0,(iext-i)*jext+1,jext,pix,irc)
  end do
  
  return

end SUBROUTINE MATWTIR

!=======================================================================

!*OPNINP   -- Open input image and get parameters
!+ 
SUBROUTINE OPNINP (iarg,im,mode,image,title,iext,jext,irc)
  character*(*) image, title
!
!     iarg: position in the command-line arguments
!     mode: file access mode: 0 = readonly, 1 = readwrite 
!--
!     author: M.Hamabe   1990-??-??
!     update: M.Hamabe   1991-08-17  (add argument "iarg")
!     update: M.Hamabe   1991-09-10/11  (check data-type)
!     update: M.Hamabe   1993-01-04  (case when command line is not valid)
!     update: M.Hamabe   1996-08-25  (support FITS)
!     update: M.Hamabe   1996-11-13  (bugfix)
!     update: M.Hamabe   1998-01-16  (assume suffix .fits, if exists)
!     update: M.Hamabe   1999-07-23  (omit IMFORT)
!----------------------------------------------------------------------

  character  :: comment*40, title0*80
  integer    :: im, naxis, irc
  integer    :: bitpix, block, naxes(2), pcount, gcount
  logical    :: simple, extendf, there
  
  if ( iarg/=0 ) then
     call CLARGC (iarg, image, irc)
     if ( irc/=0 ) then
        print '(''Enter         input  image name: '',$)'
        read '(a)', image
        irc=0
     end if
  end if
  
  if ( index(image,'.')==0 ) then
     lenf=len_trim(image)
     comment=image(:lenf)//'.fits'
     inquire (file=comment(:lenf+5), exist=there)
     if ( there ) then
        image=comment(:lenf+5)
        go to 1
     end if
     comment=image(:lenf)//'.fits.gz'
     inquire (file=comment(:lenf+8), exist=there)
     if ( there ) then
        image=comment(:lenf+8)
        go to 1
     end if
     comment=image(:lenf)//'.fit'
     inquire (file=comment(:lenf+4), exist=there)
     if ( there ) then
        image=comment(:lenf+4)
        go to 1
     end if
     comment=image(:lenf)//'.fit.gz'
     inquire (file=comment(:lenf+7), exist=there)
     if ( there ) then
        image=comment(:lenf+7)
        go to 1
     end if
     comment=image(:lenf)//'.imh'
     inquire (file=comment(:lenf+4), exist=there)
     if ( there ) then
        image=comment(:lenf+4)
        go to 1
     end if
  end if
1 continue
  lenf=len_trim(image)
  if ( index(image,'/')==0 ) then
     comment='./'//image(:lenf)
     image=comment(:lenf+2)
  end if
  print '('' Input file: '',a)', image(:len_trim(image))
  
  comment='                                      '
  
  call FTGIOU(im,irc)
  if ( irc>0 ) return
  call FTOPEN(im,image(:len_trim(image)),mode,block,irc)
  if ( irc>0 ) return
  !     read the required primary array keywords
  maxdim=2
  call FTGHPR(im,maxdim,simple,bitpix,naxis,naxes,pcount,gcount,extendf,irc)
  if ( irc>0 ) return
  iext=naxes(1)
  jext=naxes(2)
  call FTGKEY(im,'OBJECT',TITLE0,comment,irc)
  if ( irc/=0 ) then
     title0='Unknown'
  end if
  if ( irc==0 ) then
     lquote=index(title0(2:),char(39))
     title=title0(2:lquote)
  else
     title=title0
     irc=0
  end if
  print '(''     Object :'',a)', title
  print '(''     Image size ='',i4,'' x'',i4)', iext, jext
  print '(''     Bitpix     ='',i4)', bitpix
  
  return
  
end subroutine opninp

!=======================================================================

!*OPNIAF   -- Open an image-assciate file
!+
SUBROUTINE OPNIAF(IMAGE,EXT,FILE,LOUT,IRC)
  character*(*) image, ext, file
  character ::  yn
!
! OPNIAF:
!
! Arguments:
!           image  : name of the input image file
!           lout   : logical unit number of the new file
!           ext    : file-name extension of the new file
!           irc    : return code (irc=0 is normal end)
!
!--
!     AUTHOR: M. Hamabe  1995-07-08
!     update: M.Hamabe   2001-08-14 (overwrite)
! ---------------------------------------------------------------------


  len=len_trim(image)
  
  file = image(:len)//'.'//ext
  
  len=len_trim(file)
  open(LOUT,file=file, status='NEW', iostat=irc)
  if ( irc/=0 ) then
     print *
     print '(''##### ERROR: File already exists! ... '',a)', file(1:len)
     print '(''      May I overwrite? (y/<n>) '',$)'
     read  '(a)', yn
     if ( yn=='y' ) open(LOUT,file=file, iostat=irc)
     print *
  end if
  
  return

end SUBROUTINE OPNIAF

!=======================================================================

!*PUTEMSG  -- display error messages from CFITSIO
!+
SUBROUTINE PUTEMSG(irc)
!     
! PUTEMSG:
!  PUTEMSG: Display error message from CFITSIO libraries
!
!         CALL PUTEMSG(irc)
!
! Arguments:
!           irc    : Return code from CFITSIO 
!--
!     AUTHOR: M. Hamabe  1996-08-28
!     UPDATE: M. Hamabe  1999-07-17
! ---------------------------------------------------------------------

  character :: errmsg*80
  
  call FTGERR (irc, errmsg)
  
  if ( irc/=0 ) print '('' Error: '',i5,'' '',a80)', irc, errmsg
  
  return

end SUBROUTINE PUTEMSG

!=======================================================================

!*FCLOSE   -- close a file opened by FITSIO
!+
SUBROUTINE FCLOSE(im,irc)

! FCLOSE:
!
!--
!     AUTHOR: M. Hamabe  1996-08-28
!     UPDATE: M. Hamabe  1999-07-17
! ---------------------------------------------------------------------

  call FTCLOS(im,irc)
  
  return

end SUBROUTINE FCLOSE

!=======================================================================

!*GETKEYE  -- get FITS-header keyword value (real)
!+
SUBROUTINE GETKEYE(im,keyword,value,irc)
  character*(*)    keyword

! GETKEYE:
!         get FITS-header keyword value
! Arguments:
!           im     : logical unit number of data file
!           keyword: header keyword
!           irc    : Return code from FITSIO 
!           value  : keyword value
!
!--
!     AUTHOR: M. Hamabe  1996-08-28
!     UPDATE: M. Hamabe  1996-09-02
!     UPDATE: M. Hamabe  1999-07-17
!     UPDATE: M. Hamabe  2000-07-24 (bugfix)
! ---------------------------------------------------------------------

  character ::   comments*70
  
  irc=0
  call FTGKYE(im, keyword, value, comments, irc)
  if ( irc/=0 ) then
     irc=0
     call FTGKYJ(im, keyword, ivalue, comments, irc)
     value=ivalue
  end if
  
  if ( irc/=0 ) print '(a,a,a)', 'Warning: keyword ',keyword,' not found'
  
  return

end SUBROUTINE GETKEYE

!=======================================================================

!*GETKEYS  -- get FITS-header keyword value (character)
!+
SUBROUTINE GETKEYS(im,keyword,string,irc)
  character*(*) keyword, string

! GETKEYS:
!         get FITS-header keyword value
!
!         character keyword*8, string*??
!         CALL GETKEYS(im, keyword, string, irc)
!
! Arguments:
!           im     : logical unit number of data file
!           keyword: header keyword
!           irc    : Return code from FITSIO 
!           string : keyword value
!
!--
!     AUTHOR: M. Hamabe  1996-08-28
!     UPDATE: M. Hamabe  1996-09-02
!     UPDATE: M. Hamabe  1999-07-17
!     UPDATE: M. Hamabe  2000-07-24 (bugfix)
! ---------------------------------------------------------------------
  character :: comments*70
  
  irc=0
  call FTGKYS(im, keyword, string, comments, irc)
  
  if ( irc/=0 ) print '(a,a,a)', 'Warning: keyword ',keyword,' not found'

  return

end SUBROUTINE GETKEYS

!=======================================================================

!*GETKEYJ  -- get FITS-header keyword value (integer)
!+
SUBROUTINE GETKEYJ(im,keyword,number,irc)
  character*(*)   keyword
!
! GETKEYJ:
!
! Arguments:
!           im     : logical unit number of data file
!           keyword: header keyword
!           irc    : Return code from FITSIO 
!           number : keyword value
!
!--
!     AUTHOR: M. Hamabe  1996-08-28
!     UPDATE: M. Hamabe  1996-09-02
!     UPDATE: M. Hamabe  1999-07-17
!     UPDATE: M. Hamabe  2000-07-24 (bugfix)
! ---------------------------------------------------------------------

  character  :: comments*70
  
  irc=0
  
  call FTGKYJ(im, keyword, number, comments, irc)
  if ( irc/=0 ) then
     irc=0
     call FTGKYE(im, keyword, val, comments, irc)
     number=int(val+0.5)
  end if
  
  if ( irc/=0 ) print '(a,a,a)', 'Warning: keyword ',keyword,' not found'
  
  return

end SUBROUTINE GETKEYJ

!=======================================================================

!*PUTKEYE  -- put FITS-header keyword (real)
!+
SUBROUTINE PUTKEYE(im,keyword,value,decimal,comments,irc)
  character*(*) keyword, comments
  integer  ::   decimal
! 
! PUTKEYE:
!
! Arguments:
!           im     : logical unit number of data file
!           keyword: header keyword
!           irc    : Return code from IMFORT & FITSIO 
!           value  : keyword value
!
!--
!     AUTHOR: M. Hamabe  1996-08-31
!     UPDATE: M. Hamabe  1999-07-17
!     UPDATE: M. Hamabe  2000-07-24 (bugfix)
! ---------------------------------------------------------------------

  irc=0
  
  call FTPKYF(im, keyword, value, decimal, comments, irc)
  
  if ( irc/=0 ) print '(a,a)', 'Keyword:',keyword
  
  return

end SUBROUTINE PUTKEYE

!=======================================================================

!*PUTKEYS  -- Put FITS keyword (character)
!+
SUBROUTINE PUTKEYS(im,keyword,string,comments,irc)
  character*(*) keyword, string, comments
!
! Arguments:
!           im     : logical unit number of data file
!           keyword: header keyword
!           irc    : Return code from FITSIO 
!           string : keyword value
!
!--
!     AUTHOR: M. Hamabe  1996-08-31
!     UPDATE: M. Hamabe  1999-07-17
!     UPDATE: M. Hamabe  2000-07-24 (bugfix)
!C ---------------------------------------------------------------------

  irc=0
  
  call FTPKYS(im, keyword, string, comments, irc)
  
  if ( irc/=0 ) print '(a,a)', 'Keyword:',keyword
  
  return

end SUBROUTINE PUTKEYS

!=======================================================================

!*PUTKEYJ  -- Put FITS keyword (integer)
!+
SUBROUTINE PUTKEYJ(im,keyword,number,comments,irc)
  character*(*) keyword, comments
!
! Arguments:
!           im     : logical unit number of data file
!           keyword: header keyword
!           irc    : Return code from FITSIO 
!           number : keyword value
!
!--
!     AUTHOR: M. Hamabe  1996-08-31
!     UPDATE: M. Hamabe  1999-07-17
!     UPDATE: M. Hamabe  2000-07-24 (bugfix)
!C ---------------------------------------------------------------------

  irc=0
  
  call FTPKYJ(im, keyword, number, comments, irc)
  
  if ( irc/=0 ) print '(a,a)', 'Keyword:',keyword
  
  return

end SUBROUTINE PUTKEYJ

!=======================================================================

!*IMGCOPY  -- Create a copy of FITS image
!+
SUBROUTINE IMGCOPY(im,nim,bitpix,image,irc)
  character*(*) image
  integer    ::   bitpix

! Arguments:
!           im     : logical unit number of old data file
!           nim    : logical unit number of new data file
!           bitpix : bit/pix of the output FITS image
!           image  : name of the output FITS image
!           irc    : Return code from FITSIO 
!--
!     AUTHOR: M. Hamabe  1996-08-31
!     UPDATE: M. Hamabe  1999-07-22
! ---------------------------------------------------------------------
  character :: extend*80
  
  call FTGIOU(nim,irc)
  iblksize=1
  call FTINIT(nim,extend(image,'fits'),iblksize,irc)
  morekeys=20
  call FTCOPY(im,nim,morekeys,irc)
  call FTUKYJ(nim,'BITPIX',bitpix,' ',irc)
  call FTRDEF(nim,irc)
  
  return

end SUBROUTINE IMGCOPY

!=======================================================================

!*IMGNEW   -- Create a new FITS image
!+
SUBROUTINE IMGNEW(IM,IMAGE,iext,jext,BITPIX,IRC)
  character*(*)  IMAGE
  integer    ::  BITPIX
!
! Arguments:
!           im     : logical unit number of data file
!           image  : name of the output FITS image
!           iext   : size of the image
!           jext   : size of the image
!           bitpix : bit/pix of the output FITS image
!           irc    : Return code from FITSIO 
!           number : keyword value
!
!--
!     AUTHOR: M. Hamabe  1996-08-29
!     UPDATE: M. Hamabe  1999-07-17
! ---------------------------------------------------------------------

  character  ::  cdum*80
  integer    ::  pcount, gcount, naxes(2), laxis(7)
  logical    ::  simple, extend
  
  laxis(1)=iext
  laxis(2)=jext
  naxis=2
  lfn=index(image,' ')-1
  
  call FTGIOU(im,irc)
  if ( irc/=0 ) return
  iblksize=1
  cdum=image(1:lfn)//'.fits'
  call FTINIT(im,cdum,iblksize,irc)
  if ( irc/=0 ) return
  simple=.true.
  !     bitpix=-32
  naxis=2
  naxes(1)=laxis(1)
  naxes(2)=laxis(2)
  pcount=0
  gcount=1
  extend=.FALSE.
  call FTPHPR(im,simple,bitpix,naxis,naxes,pcount,gcount,extend,irc)
  if ( irc/=0 ) return
  
  return

end SUBROUTINE IMGNEW

!=======================================================================

!*ADHSTRY  -- ADD SPIRAL'S HISTORY TO THE IMAGE-HEADER
!+
SUBROUTINE ADHSTRY(im,prg_name,lprg,prg_comm,irc)
  character prg_name*(*), prg_comm*(*)

! Arguments:
!           IM       : IMAGE DESCRIPTOR 
!           PRG_NAME : NAME OF THE PROGRAM
!           LPRG     : LENGTH OF THE STRING 'PRG_NAME'
!                      If LPRG=0, then NO program name will be recorded
!           PRG_COMM : COMMENT FOR THE PROCESS BY PRG_NAME (A50)
!                      must be less than 50 characters.
!           IRC      : RETURN CODE (IRC=0 : NO ERROR)
!
!--
!     AUTHOR: M.Hamabe  1994-09-19
!     UPDATE: M.Hamabe  1996-08-21 (support FITS format)
!     UPDATE: M.Hamabe  1996-09-02
!     UPDATE: M.Hamabe  1999-07-17
! ---------------------------------------------------------------------
  character :: cdate*8, keyw*8, pname*9, svalue*70, cvalue*70, comments*70
  
  irc=0
  if ( lprg>0 ) then
     call getdate(cdate)
     pname='         '
     if ( lprg<=8 ) then
        pname(1:lprg) = prg_name(1:lprg)
     else
        pname(1:7)    = prg_name(1:7)
        pname(8:8)    = prg_name(lprg:lprg)
     end if
     pname(9:9)='@'
  end if
  keyw='SPIRL000'
  do i=1, 999
     if ( i<=9 ) then
        write(keyw(8:8),'(i1)') i
     else if ( i<=99 ) then
        write(keyw(7:8),'(i2)') i
     else
        write(keyw(6:8),'(i3)') i
     end if
     call FTGKYS(im,keyw,cvalue,comments,irc)
     if ( irc/=0 ) then
        irc=0
        if ( lprg>0 ) then
           svalue=pname//cdate//' '//prg_comm
        else
           svalue='        '//prg_comm
        end if
        call FTPKYS(im,keyw,svalue,' ',irc)
        irc=0
        return
     end if
  end do
  
  irc=1
  print *, 'Warning: History area for SPIRAL was overfllowed'
  print *, 'Warning: History was not recorded!'
  
  return

end subroutine adhstry

      
!     IRAF-compatible subroutines CLARG[C,I,R,D]
      
!---- Error code for CLARG routine -----------------------------------------
!     IE_GCMDLN    (12): cannot read command line string
!     IE_NEXARG    (14): nonexistent command line argument referenced
!     IE_NONNUMARG (16): command line argument cannot be decoded as a number
!----------------------------------------------------------------------------
!=======================================================================

!*CLARGC   -- Read command line argument as a character string.
!+
SUBROUTINE CLARGC(iarg,c_str,irc)
  
  character*(*) c_str
  
  irc=0
  if ( iargc()==0 ) then
     irc=12
     c_str=''
     return
  else if ( iarg>iargc() .or. iarg<1 ) then
     irc=14
     c_str=''
     return
  end if
  call getarg(iarg,c_str)
  
  return

end subroutine clargc

!----------------------------------------------------------------------------

SUBROUTINE CLARGI(iarg,i_number,irc)
  
  character :: cnum*20
  
  irc=0
  if ( iargc()==0 ) then
     irc=12
     i_number=0
     return
  else if ( iarg>iargc() .or. iarg<1 ) then
     irc=14
     i_number=0
     return
  end if
  call getarg(iarg,cnum)
  read (cnum,*,err=1) i_number
  return
  
1 irc=16
  i_number=0
  
  return

end subroutine clargi

!----------------------------------------------------------------------------
SUBROUTINE CLARGR(iarg,s_number,irc)
  
  character :: cnum*20
  
  irc=0
  if ( iargc()==0 ) then
     irc=12
     s_number=0
     return
  else if ( iarg>iargc() .or. iarg<1 ) then
     irc=14
     s_number=0.0
     return
  end if
  call getarg(iarg,cnum)
  read (cnum,*,err=1) s_number
  return
  
1 irc=16
  s_number=0
  
  return

end subroutine clargr
!----------------------------------------------------------------------------
SUBROUTINE CLARGD(iarg,d_number,irc)
  
  character :: cnum*20
  real(8)   :: d_number
  
  irc=0
  if ( iargc()==0 ) then
     irc=12
     d_number=0
     return
  else if ( iarg>iargc() .or. iarg<1 ) then
     irc=14
     d_number=0.0D0
     return
  end if
  call getarg(iarg,cnum)
  read (cnum,*,err=1) d_number
  return
  
1 irc=16
  d_number=0
  
  return

end subroutine clargd
